#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

#include "corpus.h"

#define MEMBLOCK 1024

guint32  corpus_sentences_nr_from_index(char *filename) {
    gchar *file = g_strdup_printf("%s.index", filename);    
    guint32 x;
    FILE *fd = fopen(file, "r");
    g_free(file);
    if (!fd) return 0;

    fread(&x, sizeof(guint32), 1, fd);
    fclose(fd);
    return x;
}


Corpus *corpus_new(void)
{
    Corpus* corpus;
    corpus = g_new(Corpus, 1);
    if (!corpus) return NULL;

    corpus->length = MEMBLOCK;
    corpus->readptr = 0;
    corpus->addptr = 0;

    corpus->words = g_new0(CorpusCell, corpus->length);
    if (!corpus->words) {
        g_free(corpus);
        return NULL;
    }

    corpus->index_size = MEMBLOCK;

    corpus->index = g_new(guint32, MEMBLOCK);
    if (!corpus->index) {
	free(corpus->words);
	g_free(corpus);
	return NULL;
    }

    corpus->index_addptr = 0;
    corpus->index[0] = 0;

    return corpus;
}

void corpus_free(Corpus *corpus)
{
    g_free(corpus->words);
    g_free(corpus->index);
    g_free(corpus);
}

int index_enlarge(Corpus *corpus)
{
    corpus->index_size += MEMBLOCK;
    corpus->index = g_renew(guint32, corpus->index, corpus->index_size);
    return (corpus->index == NULL);
}

int corpus_enlarge(Corpus *corpus)
{
    corpus->length += MEMBLOCK;
    corpus->words = g_renew(CorpusCell, corpus->words, corpus->length);
    return (corpus->words == NULL);
}

int corpus_add_word(Corpus *corpus, guint32 word, gint flags)
{
    if (corpus->addptr >= corpus->length)
        if (corpus_enlarge(corpus))
            return 1;

    if (flags == 2) {
	(corpus->words[corpus->addptr]).flags |= 0x2;
    } else if (flags == 3) {
	(corpus->words[corpus->addptr]).flags |= 0x1;
    }
    
    (corpus->words[corpus->addptr++]).word = word;

    if (!word) {
	corpus->index_addptr++;
	if (corpus->index_addptr == corpus->index_size)
	    if (index_enlarge(corpus))
		return 1;
	corpus->index[corpus->index_addptr] = corpus->addptr;
    }

    return 0;
}

CorpusCell* corpus_first_sentence(Corpus *corpus)
{
    corpus->readptr = 0;
    return corpus->words;
}

CorpusCell* corpus_next_sentence(Corpus *corpus)
{
    CorpusCell *w;
    guint32 p, l;
    p = corpus->readptr;
    l = corpus->addptr;
    w = corpus->words;
    while (p<=l && w[p].word) p++;
    if (p >= l-1) {
        corpus->readptr = 0;
        return NULL;
    }
    else
        return (w + (corpus->readptr = ++p));
}

int corpus_sentence_length(CorpusCell *s)
{
    int l = 0;
    while ((*s).word) { l++; s++; }
    return l;
}

int corpus_load(Corpus *corpus, char *filename)
{
    FILE *fd;
    long len;
    fd = gzopen(filename, "rb");
    if (fd == NULL)
        return 1;
    if (gzread(fd, &len, sizeof(guint32)) != sizeof(guint32)) return 1;

    corpus->words = g_new(CorpusCell, len);
    if (corpus->words == NULL)
        return 3;
    if (gzread(fd, corpus->words, sizeof(CorpusCell)*len) != sizeof(CorpusCell)*len)
        return 4;
    if (gzclose(fd))
        return 5;
    corpus->length = len;
    corpus->addptr = len;
    corpus->readptr = 0;

    corpus->index = NULL;
    corpus->index_size = 0;
    corpus->index_addptr = 0;
    return 0;
}

int corpus_save(Corpus *corpus, char *filename)
{
    FILE *fd;
    char *file;
    long len;
    fd = gzopen(filename, "wb");
    if (fd == NULL)
	return 1;
    len = corpus->addptr;
    if (gzwrite(fd, &corpus->addptr, sizeof(guint32)) != sizeof(guint32))
	return 1;
    if (gzwrite(fd, corpus->words, sizeof(CorpusCell) * len) != len*sizeof(CorpusCell))
	return 1;
    if (gzclose(fd))
	return 1;

    file = g_strdup_printf("%s.index", filename);
    
    fd = fopen(file, "wb");
    if (fd == NULL) 
	return 1;
    if (!fwrite(&(corpus->index_size), sizeof(guint32), 1, fd))
	return 0;
    if (!fwrite(corpus->index, sizeof(guint32), corpus->index_size, fd))
	return 0;
    if (fclose(fd))
	return 1;

    g_free(file);

    return 0;
}

guint32 corpus_diff_words_nr(Corpus *corpus)
{
    CorpusCell *w;
    guint32 max;
    guint32 p,l;
    max = 0;
    p = 0;
    l = corpus->addptr;
    w = corpus->words;
    while (p < l) {
        if ((*w).word > max) max = (*w).word;
        p++;
        w++;
    }
    return max;
}

guint32 corpus_sentences_nr(Corpus *corpus)
{
    guint32 result;
    CorpusCell *s;
    result = 0;

    if (!corpus->addptr) return 0;

    s = corpus_first_sentence(corpus);
    while (s != NULL) {
        result++;
        s = corpus_next_sentence(corpus);
    }
    return result;
}



char *capital_dup(const char* str) {
    char *capital = g_strdup(str);
    capital[0] = g_ascii_toupper(capital[0]);
    return capital;
}

char *uppercase_dup(const char* str) {
    char *uppercase = g_strdup(str);
    char *ptr = uppercase;
    while(*ptr) {
	*ptr = g_ascii_toupper(*ptr);
	ptr++;
    }
    return uppercase;
}

char *lowercase(char* str) {
    char *ptr = str;
    while(*ptr) {
	if (g_ascii_isupper(*ptr)) {
	    *ptr = g_ascii_tolower(*ptr);
	}
	ptr++;
    }
    return str;
}

int isCapital(const char* str) {
    int i = 0;
    if (g_ascii_isupper(str[i])) {
	int ok = 1;
	i++;
	while(ok && str[i]) {
	    if (!g_ascii_islower(str[i])) ok = 0;
	    i++;
	}
	return ok;
    } else {
	return 0;
    }
}

int isUPPERCASE(const char* str) {
    int i = 0;
    int ok = 1;
    while(ok && str[i]) {
	if (!g_ascii_isupper(str[i])) ok = 0;
	i++;
    }
    return ok;
}
